@@ -7,9 +7,11 @@
  * of course I am not responsible for any damages caused by
  * this assembler generating bad code or anything like that)
  *
- * Copyright 1995-1996 by Timo Rossi
- *   email: trossi@cc.jyu.fi
- *   www: http://www.jyu.fi/~trossi/
+ * Copyright 1995-1998 by Timo Rossi
+ * See the file picasm.doc for more information
+ * 
+ *   email: trossi@iki.fi
+ *   www: http://www.iki.fi/trossi/pic/
  *
  */
 
@@ -22,7 +24,7 @@
 
 #include "picasm.h"
 
-int warnmode;
+int warnlevel;
 
 static int total_line_count;
 static int errors, warnings; /* error & warning counts */
@@ -34,7 +36,8 @@ extern struct pic_type pic_types[];
 
 static FILE *list_fp;
 static int listing_on;
-static unsigned short list_loc, *list_ptr;
+static int list_loc;
+static pic_instr_t *list_ptr;
 static long list_val, list_len;
 
 int cond_nest_count;
@@ -52,17 +55,18 @@ struct patch **local_patch_list_ptr;
 
 struct pic_type *pic_type;
 
-unsigned short instr_set;
-int prog_mem_size, reg_file_size;
+int prog_mem_size;
+
+static int reg_file_limit;
 
 short code_generated;
 
-static unsigned short prog_mem[PROGMEM_MAX];
-static unsigned short data_mem[DATAMEM_MAX];
-static unsigned short pic_id[4];
-unsigned short config_fuses;
+static pic_instr_t prog_mem[PROGMEM_MAX];
+static pic_instr_t data_eeprom[EEPROM_MAX];
+static pic_instr_t pic_id[4];
+pic_instr_t config_fuses;
 
-int O_Mode;
+org_mode_t O_Mode;
 
 int prog_location;         /* current address for program code */
 int reg_location;   /* current address for register file */
@@ -83,7 +87,8 @@ err_line_ref(void)
   if(current_file != NULL) {
     inc = current_file;
     if(inc->type != INC_FILE) {
-      fprintf(stderr, "(Macro %s) ", inc->v.m.sym->name);
+      fprintf(stderr, "(Macro %s line %d) ",
+	      inc->v.m.sym->name, inc->linenum);
       while(inc != NULL && inc->type != INC_FILE)
 	inc = inc->next;
     }
@@ -149,8 +154,9 @@ error(int lskip, char *fmt, ...)
   }
   fputc('\n', stderr);
   va_end(args);
-  errors++;
-
+  if(++errors >= MAX_ERRORS)
+    fatal_error("too many errors, aborting");
+	
   if(lskip)
     error_lineskip();
 }
@@ -190,7 +196,7 @@ mem_alloc(int size)
 static void
 init_assembler(void)
 {
-  unsigned short *cp;
+  pic_instr_t *cp;
   int n;
 
   init_symtab(); /* initialize symbol table */
@@ -199,7 +205,7 @@ init_assembler(void)
   for(n = PROGMEM_MAX, cp = prog_mem; n-- > 0; *cp++ = INVALID_INSTR);
 
   /* initialize data memory to invalid values */
-  for(n = DATAMEM_MAX, cp = data_mem; n-- > 0; *cp++ = INVALID_DATA);
+  for(n = EEPROM_MAX, cp = data_eeprom; n-- > 0; *cp++ = INVALID_DATA);
 
   /* initialize jump  patch list */
   global_patch_list = NULL;
@@ -227,6 +233,8 @@ init_assembler(void)
   code_generated = 0;
 
   local_level = 0;
+
+  ifskip_mode = 0;
 }
 
 /*
@@ -259,7 +267,7 @@ void gen_code(int val)
   if(prog_mem[prog_location] != INVALID_INSTR)
     warning("Overlapping code at 0x%x\n", prog_location);
 
-  if((list_flags & LIST_LOC) == 0) {
+  if((list_flags & LIST_PROG) == 0) {
     list_loc = prog_location;
     list_flags = LIST_LOC | LIST_PROG;
   }
@@ -292,10 +300,10 @@ gen_edata(int val)
     return;
   }
 
-  if(edata_location >= pic_type->datamem_size)
+  if(edata_location >= pic_type->eeprom_size)
     fatal_error("Data EEPROM address out of range");
 
-  if(data_mem[edata_location] < 0x100)
+  if(data_eeprom[edata_location] < 0x100)
     warning("Overlapping EEPROM data at 0x%x\n", edata_location);
 
   if((list_flags & LIST_LOC) == 0) {
@@ -304,7 +312,7 @@ gen_edata(int val)
   }
   list_len++;
 
-  data_mem[edata_location++] = val;
+  data_eeprom[edata_location++] = val;
 
   code_generated = 1;
 }
@@ -316,7 +324,7 @@ static void
 write_hex_record(FILE *fp,
 			   int reclen, /* length (in words) */
 			   int loc, /* address */
-			   unsigned short *data, /* pointer to word data */
+			   pic_instr_t *data, /* pointer to word data */
 			   int format) /* IHX8M or IHX16 */
 {
   int check = 0;
@@ -336,11 +344,13 @@ write_hex_record(FILE *fp,
   while(reclen--) {
     switch(format) {
       case IHX8M:
-        fprintf(fp, "%02X%02X", *data & 0xff, (*data >> 8) & 0xff);
+        fprintf(fp, "%02X%02X",
+		(int)(*data & 0xff), (int)((*data >> 8) & 0xff));
 	break;
 
       case IHX16:
-	fprintf(fp, "%02X%02X", (*data >> 8) & 0xff, *data & 0xff);
+	fprintf(fp, "%02X%02X",
+		(int)((*data >> 8) & 0xff), (int)(*data & 0xff));
 	break;
     }
     check += (*data & 0xff) + ((*data >> 8) & 0xff);
@@ -348,7 +358,7 @@ write_hex_record(FILE *fp,
   }
 
   /* write checksum, assumes 2-complement */
-  fprintf(fp, "%02X\n", (-check) & 0xff);
+  fprintf(fp, "%02X" HEX_EOL, (-check) & 0xff);
 }
 
 /*
@@ -383,7 +393,7 @@ write_output(char *fname, int format)
 
   /* PIC ID */
   if(pic_id[0] != INVALID_ID) {
-    switch(instr_set) {
+    switch(pic_type->instr_set) {
       case PIC12BIT:
         write_hex_record(fp, 4, prog_mem_size, pic_id, format);
 	break;
@@ -396,37 +406,29 @@ write_output(char *fname, int format)
 
   /* config fuses */
   if(config_fuses != INVALID_CONFIG) {
-    switch(instr_set) {
-      case PIC12BIT:
-        write_hex_record(fp, 1, 0xfff, &config_fuses, format);
-	break;
-
-      case PIC14BIT:
-	write_hex_record(fp, 1, 0x2007, &config_fuses, format);
-	break;
-    }
+    write_hex_record(fp, 1, pic_type->fuse_addr, &config_fuses, format);
   }
 
-  if(pic_type->datamem_size > 0) { /* data EEPROM */
+  if(pic_type->eeprom_size > 0) { /* data EEPROM */
     for(loc = 0;;) {
-      while(loc < pic_type->datamem_size && data_mem[loc] >= 0x100)
+      while(loc < pic_type->eeprom_size && data_eeprom[loc] >= 0x100)
 	loc++;
         
-      if(loc >= pic_type->datamem_size)
+      if(loc >= pic_type->eeprom_size)
 	break;
         
       reclen = 0;
-      while(reclen < 8 && loc < pic_type->datamem_size
-	    && data_mem[loc] < 0x100) {
+      while(reclen < 8 && loc < pic_type->eeprom_size
+	    && data_eeprom[loc] < 0x100) {
 	loc++;
 	reclen++;
       }
       write_hex_record(fp, reclen, 0x2100+loc-reclen,
-		       &data_mem[loc-reclen], format);
+		       &data_eeprom[loc-reclen], format);
     }
   }
 
-  fputs(":00000001FF\n", fp); /* end record */
+  fputs(":00000001FF" HEX_EOL , fp); /* end record */
   fclose(fp);
 }
 
@@ -454,15 +456,18 @@ write_listing_line(int cond_flag)
 	else
 	  fputs("    ", list_fp);
 
-	if((list_flags & (LIST_PROG|LIST_EDATA|LIST_PTR)) != 0)
-	  fprintf(list_fp, "%c%04X ",
-		  ((list_flags & LIST_FORWARD) ? '?' : ' '),
-		  ((list_flags & LIST_PROG) ?
-		   prog_mem[list_loc] :
-		   ((list_flags & LIST_EDATA) ?
-		    data_mem[list_loc] : *list_ptr++)));
-	else
+	if((list_flags & (LIST_PROG|LIST_EDATA|LIST_PTR)) != 0) {
+	  fputc(((list_flags & LIST_FORWARD) ? '?' : ' '), list_fp);
+	  if(list_flags & LIST_PROG) {
+	    fprintf(list_fp, "%04X ", prog_mem[list_loc]);
+	  } else if(list_flags & LIST_EDATA) {
+	    fprintf(list_fp, "%04X ", data_eeprom[list_loc]);
+	  } else if(list_flags & LIST_PTR) {
+	    fprintf(list_fp, "%04X ", *list_ptr++);
+	  }
+	} else {
 	  fputs("      ", list_fp);
+	}
       }
 
       fputs(line_buffer, list_fp);
@@ -484,11 +489,13 @@ write_listing_line(int cond_flag)
 	else
 	  fputs("    ", list_fp);
 
-	fprintf(list_fp," %04X\n",
-		((list_flags & LIST_PROG) ?
-		 prog_mem[list_loc] :
-		 ((list_flags & LIST_EDATA) ?
-		  data_mem[list_loc] : *list_ptr++)));
+	if(list_flags & LIST_PROG) {
+	  fprintf(list_fp, " %04X\n", prog_mem[list_loc]);
+	} else if(list_flags & LIST_EDATA) {
+	  fprintf(list_fp, " %04X\n", data_eeprom[list_loc]);
+	} else if(list_flags & LIST_PTR) {
+	  fprintf(list_fp, " %04X\n", *list_ptr++);
+	}
       }
     }
 
@@ -608,6 +615,7 @@ if_else_skip(void)
   int t, ccount;
 
   ccount = 0;
+  ifskip_mode++;
   do {
     skip_eol();
     get_token();
@@ -617,7 +625,7 @@ if_else_skip(void)
     if(token_type == TOK_IDENTIFIER) {
       t = 1;
       get_token();
-      if(token_type == TOK_COMMA)
+      if(token_type == TOK_COLON)
 	get_token();
     }
     if(token_type == KW_IF) {
@@ -631,17 +639,15 @@ if_else_skip(void)
     }
   } while(token_type != TOK_EOF && token_type != KW_END);
 
-  if(token_type == TOK_EOF || token_type == KW_END)
-    return -1;
-  else
-    return t;
+  ifskip_mode--;
+  return (token_type == TOK_EOF || token_type == KW_END) ? -1 : t;
 }
 
 /*
  * Add a patch pointing to the current location
  */
 void
-add_patch(int tab, struct symbol *sym, int type)
+add_patch(int tab, struct symbol *sym, patchtype_t type)
 {
   struct patch **patch_list_ptr, *ptch;
 
@@ -689,7 +695,7 @@ apply_patches(struct patch *patch_list)
     else
       switch(ptch->type) {
         case PATCH8:
-	  if(instr_set == PIC12BIT && (ptch->label->v.value & 0x100) != 0
+	  if(pic_type->instr_set == PIC12BIT && (ptch->label->v.value & 0x100) != 0
 	     && (prog_mem[ptch->location] & 0xff00) == 0x900)
 	    error(0, "CALL address in upper half of a page (label '%s%s')",
 		  (patch_list == global_patch_list ? "" : "="),
@@ -952,6 +958,13 @@ do_label:
     if(token_type == KW_END)
       break;
 
+    if(token_type == KW_ERROR) {
+      while(isspace((unsigned char)(*line_buf_ptr)))
+	 line_buf_ptr++;
+      error(1, "%s", line_buf_ptr);
+      continue;
+    }
+	  
     op = token_type;
     get_token();
 
@@ -1023,7 +1036,7 @@ do_label:
       apply_patches(*local_patch_list_ptr);
       remove_local_symtab();
       break;
-
+	    
     case KW_IF:
       if(sym != NULL)
 	error(0, "Label not allowed with IF");
@@ -1092,22 +1105,25 @@ do_label:
       org_val = -1;
       if((t = org_mode()) != O_NONE) {
 	get_token();
-	O_Mode = t;
+	O_Mode = (org_mode_t)t;
 	switch(O_Mode) {
 	  case O_PROGRAM:
-	    t = prog_location;
+	  case O_NONE: /* shut up GCC warning, cannot really happen here */
+	    org_val = prog_location;
 	    break;
 
 	  case O_REGFILE:
-	    t = reg_location;
+	    org_val = reg_location;
 	    break;
 
 	  case O_EDATA:
-	    t = edata_location;
+	    org_val = edata_location;
 	    break;
 	}
-	if(t < 0)
+	if(org_val < 0) {
 	  error(0, "ORG value not set");
+	  O_Mode = O_NONE;
+	}
 	break;
       }
 
@@ -1125,14 +1141,30 @@ do_label:
 
       if(token_type == TOK_COMMA) {
 	get_token();
-	if((t = org_mode()) == O_NONE)
+	if((t = org_mode()) == O_NONE) {
 	  error(0, "Invalid ORG mode");
-	else
-	  O_Mode = t;
-
+	} else {
+	  O_Mode = (org_mode_t)t;
+	  switch(O_Mode)
+	  {
+	    case O_PROGRAM:
+	    case O_NONE: /* shut up GCC warning, cannot really happen here */
+	      prog_location = org_val;
+	      break;
+		  
+	    case O_REGFILE:
+	      reg_location = org_val;
+	      break;
+		  
+	    case O_EDATA:
+	      edata_location = org_val;
+	      break;
+	  }
+	}
+	      
 	get_token();
       }
-
+	    
       list_loc = org_val;
       list_flags = LIST_LOC;
       break;
@@ -1154,7 +1186,7 @@ do_label:
       if(O_Mode != O_REGFILE)
 	error(0, "ORG mode conflict");
       else {
-	if(reg_location >= reg_file_size)
+	if(reg_location >= reg_file_limit)
 	  fatal_error("Register file address out of range");
 	list_loc = reg_location;
 	list_flags = LIST_LOC;
@@ -1166,7 +1198,7 @@ do_label:
       if(pic_type == NULL)
 	fatal_error("PIC device type not set");
 
-      if(pic_type->datamem_size == 0) {
+      if(pic_type->eeprom_size == 0) {
 	error(1, "PIC%s does not have data EEPROM", pic_type->name);
 	continue;
       }
@@ -1179,7 +1211,7 @@ do_label:
 	} else {
 	  val = get_expression();
 	  if(expr_error)
-	    continue;
+	    break;
 
 	  if(val < 0 || val > 0xff) {
 	    error(0, "Data EEPROM byte out of range");
@@ -1236,9 +1268,8 @@ do_label:
 	continue;
       }
       pic_type = pic;
-      instr_set = pic_type->instr_flags & INSTRSET_MASK;
       prog_mem_size = pic_type->progmem_size;
-      reg_file_size = pic_type->regfile_size;
+      reg_file_limit = pic_type->regfile_limit;
 
       sprintf(symname, "__%s", pic_type->name);
       sym = add_symbol(symname, SYMTAB_GLOBAL);
@@ -1275,7 +1306,7 @@ do_label:
 	    error(1, "PIC ID too long (max 4 bytes)");
 	    continue;
 	  }
-	  pic_id[t] = val;
+	  pic_id[t] = (pic_instr_t)val;
 	}
 
 	t++;
@@ -1299,7 +1330,7 @@ do_label:
       /* mnemonics */
 
     default:
-      switch(instr_set) {
+      switch(pic_type->instr_set) {
         case PIC12BIT:
 	  t = assemble_12bit_mnemonic(op);
 	  break;
@@ -1357,7 +1388,7 @@ main(int argc, char *argv[])
   pic_type = NULL;
   out_filename[0] = '\0';
   list_filename[0] = '\0';
-  warnmode = 0;
+  warnlevel = 0;
 
   while(argc > 1 && argv[1][0] == '-') {
     switch(argv[1][1]) {
@@ -1399,9 +1430,8 @@ main(int argc, char *argv[])
 	  exit(EXIT_FAILURE);
 	}
 
-	instr_set = pic_type->instr_flags & INSTRSET_MASK;
 	prog_mem_size = pic_type->progmem_size;
-	reg_file_size = pic_type->regfile_size;
+	reg_file_limit = pic_type->regfile_limit;
 	break;
 
       case 'l': /* listing/list filename */
@@ -1415,13 +1445,17 @@ main(int argc, char *argv[])
         break;
 
       case 'w': /* warning mode (gives some more warnings) */
-	warnmode = 1;
+	if(argv[1][2] != '\0') {
+	  warnlevel = atoi(&argv[1][2]);
+	} else {
+	  warnlevel = 1;
+	}
 	break;
 
       case 'v': /* version info */
 	fprintf(stderr,
 		"12/14-bit PIC assembler " VERSION
-		" -- Copyright 1995-1996 by Timo Rossi\n");
+		" -- Copyright 1995-1998 by Timo Rossi\n");
 	break;
 	
       case '-': /* end of option list */
@@ -1441,7 +1475,7 @@ opt_done:
   if(argc != 2) {
 usage:
     fputs("Usage: picasm [-o<objname>] [-l<listfile>] [-s] [-ihx8m/ihx16]\n"
-	  "              [-pic<device>] [-w] <filename>\n", stderr);
+	  "              [-pic<device>] [-w[n]] <filename>\n", stderr);
     exit(EXIT_FAILURE);
   }
 
